/*
			Copyright (c) 1994 by
			Advanced Visual Systems Inc.
			All Rights Reserved

	This software comprises unpublished confidential information of
	Advanced Visual Systems Inc. and may not be used, copied or made
	available to anyone, except in accordance with the license
	under which it is furnished.

	This file is under Perforce control
	$Id: //depot/express/fcs992/modules/pln_dist.c#1 $
*/
#include <avs/util.h>
#include <avs/err.h>
#include <avs/om.h>
#include <avs/fld.h>
#include <avs/math.h>
#include <avs/arr.h>
#include <avs/om_att.h>

#define ERR_RETURN(A) ERRerror("plane_dist",1,ERR_ORIG, A); return(0);
#define MAX_NAME_SIZE 1024
#define EPS  0.0001

static float IXFORM[16] = {1.0, 0.0, 0.0, 0.0,
			   0.0, 1.0, 0.0, 0.0,
			   0.0, 0.0, 1.0, 0.0,
			   0.0, 0.0, 0.0, 1.0};

int FUNCplane_dist (OMobj_id in, OMobj_id plane, OMobj_id out, float *dist);

int DVplane_distance_update(elem_id)
    OMobj_id elem_id;
{
	OMobj_id in, out, plane;
	float dist;

	if (OMis_null_obj(elem_id)) {
		ERR_RETURN("module id is 0");
	}

	in = OMfind_subobj(elem_id, OMstr_to_name("in"), OM_OBJ_RD);
	if (OMis_null_obj(in)) {
		ERR_RETURN("in id is 0");
	}
	out = OMfind_subobj(elem_id, OMstr_to_name("out"), OM_OBJ_RW);
	if (OMis_null_obj(out)) {
		ERR_RETURN("out id is 0");
	}
	plane = OMfind_subobj(elem_id, OMstr_to_name("plane"), OM_OBJ_RD);
	if (OMis_null_obj(plane)) {
		ERR_RETURN("plane id is 0");
	}
	if (!FUNCplane_dist(in, plane, out, &dist)) {
		return(0);
	}

	if (OMset_name_real_val (elem_id, OMstr_to_name("dist"), (double)dist) != 1) {
		ERR_RETURN("cannot set dist value");
	}
	return(1);
}

int FUNCplane_dist (OMobj_id in, OMobj_id plane, OMobj_id out, float *dist)
{
	int   nnodes, type, size, recomp;
	int   nspace, plane_nspace, npts, stat, i;
	float  *coord, *pnts;
	float  bounds[3], norm[3], norm_xfm[3], *norm_x;
	float  xfm[4][4], *in_xfm, in_xform[4][4], t_xform[4][4];
	float  cross[3], l;
	float  *node_dist, plane_dist;
	int    in_seq, out_seq;
	OMobj_id pln_id;

	/*-----------------*/
	/*   GET PLANE     */
	/*-----------------*/
	if (FLDget_nspace (plane, &plane_nspace) != 1) {
		ERR_RETURN("Error getting nspace");
	}
	if (FLDget_points (plane, &pnts, &size, OM_GET_ARRAY_RD) != 1) {
		ERR_RETURN("Error getting points");
	}
	stat = FLDget_xform(plane, (float *)xfm);
	if (stat < 0) {
		ERR_RETURN("cannot get xform for probe");
	}
	/*------------------*/
	/* TRANSFORM PLANE  */
	/*------------------*/
	stat = FLDget_xform(in, (float *)in_xform);
	if (stat < 0) {
		ERR_RETURN("cannot get xform for field");
	}
	else if (stat == 0) {
		in_xfm = (float *)0;
	}
	else if (MATmat_is_identity((float *)in_xform, 4))
		in_xfm = (float *)0;
	else
		in_xfm = (float *)in_xform;

	for (i=0; i<plane_nspace; i++)
		bounds[i] = pnts[i];

	ARRfree(pnts);

	for (; i<3; i++)
		bounds[i] = 0.0;
	norm[0] = 0.0; norm[1] = 0.0; norm[2] = 1.0;

	MATvec3_mat4_multiply(bounds, xfm);

	MATxform_vecs(1, (Matr3 *)norm, xfm, (Matr3 *)norm_xfm);

	l = VEC_LEN(norm_xfm);
	VEC_NORMALIZE(norm_xfm,l);

	plane_dist = VEC_DOT(bounds, norm_xfm);

	*dist = plane_dist;

	stat = FLDget_array_float (out, "plane_norm", &norm_x, &size, OM_GET_ARRAY_RD);
	if (stat < 1) {
		if (FLDadd_float_array(out, "plane_norm", 3) != 1) {
			ERR_RETURN("cannot create normal array");
		}
		/*
		 * Prevent this array from being saved.
		 * Eventually we might be able to determine this
		 * automatically but for now, it must be set by hand.
		 */
		pln_id = OMfind_subobj(out, OMstr_to_name("plane_norm"),
					OM_OBJ_RW);
		OMset_obj_atts(pln_id, OM_atts_nosave);
		norm_x = NULL;
	}
	out_seq = FLDget_subelem_seq(out, "Node_Data");
	in_seq = FLDget_subelem_seq(in, "Mesh");
	if (in_seq > out_seq)
		recomp = 1;
	else {
		if (stat < 1)
			recomp = 1;
		else {
			VEC_CROSS(cross, norm_xfm, norm_x);
			l = VEC_LEN(cross);
			if (l < EPS)
				recomp = 0;
			else
				recomp = 1;
		}
	}
	if (recomp) {
		if (FLDget_nnodes(in, &nnodes) != 1) {
			ERR_RETURN("cannot get nnodes");
		}
		if (FLDget_nspace(in, &nspace) != 1) {
			ERR_RETURN("cannot get nnodes");
		}
		if (FLDget_coord(in, &coord, &size, OM_GET_ARRAY_RD) != 1) {
			ERR_RETURN("cannot get coordinates");
		}

		if (FLDset_nnodes (out, nnodes) != 1) {
			ERR_RETURN("Error setting nnodes");
		}
		type = OM_TYPE_FLOAT;
		if (FLDget_node_data(out, 0, &type, (char **)&node_dist,
				      &size, OM_GET_ARRAY_WR) != 1) {
			ERR_RETURN("cannot get node data");
		}
		cross[0]=cross[1]=cross[2]=0.0;
		for (i=0; i<nnodes; i++) {
			memcpy(cross, coord+nspace*i, nspace*sizeof(float));
			if (in_xfm)
				MATvec3_mat4_multiply(cross, (Matr4 *)in_xfm);
			node_dist[i] = VEC_DOT(cross, norm_xfm);
		}
		if (FLDset_array_float (out, "plane_norm", 3, norm_xfm, OM_SET_ARRAY_COPY) != 1) {
			ERR_RETURN("cannot set normal array");
		}
		ARRfree(coord);
		ARRfree(node_dist);
	}
	if (norm_x)
	    ARRfree(norm_x);
	return(1);
}
